//    FastHexaViewer - A fast binary file viewer
//    Copyright (C) 2006  Regis COSNIER
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace FastHexaViewer
{
	public partial class MainForm : Form
	{
		Stream fileStream = null;
		int numberOfDisplayLine = 32;
		int numberOfDisplayColumn = 16;
		byte[] buffer = new byte[0];
		int trueBufferLength = 0;
		long firstCharacterAbsoluteOffset = 0;
		long lastCharacterAbsoluteOffset = 0;
		int numberOfAddressDigit = 12;
		string addressFormat;
		string binaryFormat;
		string asciiFormat;
		Thread searchTread;
		bool continueSearch = false;
		long searchResult = -1;
		string originalTitle = "";
		Encoding encoding = System.Text.Encoding.Default;

		GotoForm gotoForm = new GotoForm();
		SearchForm searchForm = new SearchForm();
		WaitForm waitForm = new WaitForm();

		private bool ContinueSearch
		{
			get { lock (this) { return this.continueSearch; } }
			set { lock (this) { this.continueSearch = value; } }
		}

		public MainForm()
		{
			InitializeComponent();

			this.originalTitle = Text;

			this.waitForm.Stopped += new EventHandler(waitForm_Stopped);
			this.textBoxHexa.MouseWheel += new MouseEventHandler(textBoxHexa_MouseWheel);

			this.addressFormat = "{0:X" + numberOfAddressDigit + "}h: ";
			this.binaryFormat = "{0:X2} ";
			this.asciiFormat = "{0}";
		}

		private void MainForm_Load(object sender, EventArgs e)
		{
			UpdateTheNumberOfDisplayLine();

			string[] options = Environment.GetCommandLineArgs();
			if (options.Length > 1)
			{
				try
				{
					this.fileStream = new FileStream(options[1], FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
					this.firstCharacterAbsoluteOffset = 0;

					UpdateTitle(options[1]);
				}
				catch (IOException ex)
				{
					MessageBox.Show(ex.Message);
				}

				for (int i = 2; i < options.Length; i++)
					Process.Start(Application.ExecutablePath, "\"" + options[i] + "\"");
			}
			else
				UpdateTitle("");
			UpdateFileSize();
			UpdateScrollBar();
			UpdateDisplay();
			UpdateCursorInfo();
		}

		private void openToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (this.fileStream != null)
			{
				this.fileStream.Close();
				this.fileStream = null;
			}

			OpenFileDialog openFileDialog = new OpenFileDialog();
			openFileDialog.Filter = "All files (*.*)|*.*";
			openFileDialog.FilterIndex = 1;
			openFileDialog.RestoreDirectory = true;
			if (openFileDialog.ShowDialog() == DialogResult.OK)
			{
				try
				{
					this.fileStream = new FileStream(openFileDialog.FileName, FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
					this.firstCharacterAbsoluteOffset = 0;
					UpdateTitle(openFileDialog.FileName);
				}
				catch (IOException ex)
				{
					MessageBox.Show(ex.Message);
					UpdateTitle("");
				}
			}
			else
				UpdateTitle("");
			UpdateFileSize();
			UpdateScrollBar();
			UpdateDisplay();
			UpdateCursorInfo();
		}

		private void MainForm_DragDrop(object sender, DragEventArgs e)
		{
			if (e.Data.GetDataPresent(DataFormats.FileDrop))
			{
				string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);

				try
				{
					this.fileStream = new FileStream(files[0], FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
					this.firstCharacterAbsoluteOffset = 0;
					UpdateTitle(files[0]);
				}
				catch (IOException ex)
				{
					MessageBox.Show(ex.Message);
					UpdateTitle("");
				}
				UpdateFileSize();
				UpdateScrollBar();
				UpdateDisplay();
				UpdateCursorInfo();

				for (int i = 1; i < files.Length; i++)
					Process.Start(Application.ExecutablePath, "\"" + files[i] + "\"");
			}
		}

		private void MainForm_DragEnter(object sender, DragEventArgs e)
		{
			if (e.Data.GetDataPresent(DataFormats.FileDrop))
				e.Effect = DragDropEffects.Move;
			else
				e.Effect = DragDropEffects.None;
		}

		private void UpdateScrollBar()
		{
			if (this.fileStream != null)
			{
				// 64bit problem can arise here+++
				this.vScrollBarHexa.Value = (int)(this.firstCharacterAbsoluteOffset / this.numberOfDisplayColumn);
				this.vScrollBarHexa.Minimum = 0;
				this.vScrollBarHexa.Maximum = (int)(this.fileStream.Length / this.numberOfDisplayColumn);
				this.vScrollBarHexa.SmallChange = 1;
				this.vScrollBarHexa.LargeChange = this.numberOfDisplayLine;
				this.vScrollBarHexa.Enabled = true;
				// 64bit problem can arise here---
			}
			else
			{
				this.vScrollBarHexa.Maximum = 0;
				this.vScrollBarHexa.Value = 0;
				this.vScrollBarHexa.Enabled = false;
			}
		}

		private void UpdateDisplay()
		{
			if (this.fileStream == null)
			{
				this.textBoxHexa.Text = "";
				return;
			}

			int previousSelectIndex = this.textBoxHexa.SelectionStart;

			this.buffer = new byte[this.numberOfDisplayLine * this.numberOfDisplayColumn];
			this.fileStream.Position = this.firstCharacterAbsoluteOffset;
			lock (this)
			{
				this.trueBufferLength = this.fileStream.Read(this.buffer, 0, this.buffer.Length);
			}
			this.lastCharacterAbsoluteOffset = this.firstCharacterAbsoluteOffset + this.trueBufferLength;

			long lineOffset = this.firstCharacterAbsoluteOffset;
			uint currentRelativeOffset = 0;
			string text = "";
			bool endOfFileReached = false;
			for (long lineNumber = 0; lineNumber < this.numberOfDisplayLine; lineNumber++)
			{
				string line = string.Format(this.addressFormat, lineOffset);
				string lineHexa = "";
				string lineAscii = "";
				int column;
				for (column = 0; column < this.numberOfDisplayColumn; column++)
				{
					if (currentRelativeOffset >= this.trueBufferLength) { endOfFileReached = true; break; }
					byte currentByte = this.buffer[currentRelativeOffset];
					lineHexa += string.Format(this.binaryFormat, currentByte);
					//lineAscii += string.Format(this.asciiFormat, currentByte >= 32 && currentByte < 128 ? char.ConvertFromUtf32(currentByte) : ".");
					lineAscii += string.Format(this.asciiFormat, (currentByte >= 32) ? encoding.GetString(new byte[] { currentByte }) : ".");
					currentRelativeOffset++;
				}
				int padding = this.numberOfDisplayColumn - column;
				for (column = 0; column < padding; column++)
				{
					lineHexa += "   ";
					lineAscii += ' ';
				}
				text += line + lineHexa + "; " + lineAscii;

				if (endOfFileReached || lineNumber + 1 == this.numberOfDisplayLine) break;

				text += "\r\n";

				lineOffset += this.numberOfDisplayColumn;
			}
			this.textBoxHexa.SuspendLayout();
			this.textBoxHexa.Text = text;
			this.textBoxHexa.Select(previousSelectIndex, 0);
			this.textBoxHexa.ResumeLayout();
		}

		private void exitToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Close();
		}

		private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
		{
			if (this.fileStream != null)
				this.fileStream.Close();
		}

		private void vScrollBarHexa_Scroll(object sender, ScrollEventArgs e)
		{
			this.firstCharacterAbsoluteOffset = this.vScrollBarHexa.Value;
			this.firstCharacterAbsoluteOffset *= this.numberOfDisplayColumn;
			UpdateDisplay();
			UpdateCursorInfo();
		}

		private void textBoxHexa_SizeChanged(object sender, EventArgs e)
		{
			UpdateTheNumberOfDisplayLine();

			UpdateScrollBar();
			UpdateDisplay();
			UpdateCursorInfo();
		}

		private void UpdateTheNumberOfDisplayLine()
		{
			Graphics g = this.textBoxHexa.CreateGraphics();
			float fontHeight = g.MeasureString("0", this.textBoxHexa.Font).Height;
			g.Dispose();
			this.numberOfDisplayLine = this.textBoxHexa.ClientSize.Height / (int)fontHeight;
		}

		private void UpdateTitle(string filename)
		{
			if (filename.Length > 0)
				Text = this.originalTitle + " - " + filename;
			else
				Text = this.originalTitle;
		}

		private void UpdateFileSize()
		{
			if (this.fileStream != null)
				this.toolStripStatusLabelFileSize.Text = "File Size : 0x" + this.fileStream.Length.ToString("X") + ", " + this.fileStream.Length.ToString();
			else
				this.toolStripStatusLabelFileSize.Text = "No File loaded";
		}

		private void UpdateCursorInfo()
		{
			long absoluteOffset = GetAbsoluteOffsetFromCursor();
			this.toolStripStatusLabelPosition.Text = "Position: 0x" + absoluteOffset.ToString("X") + ", " + absoluteOffset.ToString();
		}

		private void textBoxHexa_KeyDown(object sender, KeyEventArgs e)
		{
			switch (e.KeyCode)
			{
				case Keys.Up:
					if (e.Modifiers == Keys.Control || this.textBoxHexa.GetLineFromCharIndex(this.textBoxHexa.SelectionStart) == 0)
					{
						this.firstCharacterAbsoluteOffset -= this.numberOfDisplayColumn;
						if (this.firstCharacterAbsoluteOffset < 0)
							this.firstCharacterAbsoluteOffset = 0;

						UpdateScrollBar();
						UpdateDisplay();
					}
					break;
				case Keys.Down:
					if (e.Modifiers == Keys.Control || this.textBoxHexa.GetLineFromCharIndex(this.textBoxHexa.SelectionStart) == this.numberOfDisplayLine - 1)
					{
						this.firstCharacterAbsoluteOffset += this.numberOfDisplayColumn;
						if (this.firstCharacterAbsoluteOffset > this.lastCharacterAbsoluteOffset)
							this.firstCharacterAbsoluteOffset = this.lastCharacterAbsoluteOffset;

						UpdateScrollBar();
						UpdateDisplay();
					}
					break;
				case Keys.PageUp:
					this.firstCharacterAbsoluteOffset -= this.numberOfDisplayLine * this.numberOfDisplayColumn;
					if (this.firstCharacterAbsoluteOffset < 0)
						this.firstCharacterAbsoluteOffset = 0;

					UpdateScrollBar();
					UpdateDisplay();
					break;
				case Keys.PageDown:
					this.firstCharacterAbsoluteOffset += this.numberOfDisplayLine * this.numberOfDisplayColumn;
					if (this.firstCharacterAbsoluteOffset > this.lastCharacterAbsoluteOffset)
						this.firstCharacterAbsoluteOffset = this.numberOfDisplayColumn * (this.lastCharacterAbsoluteOffset / this.numberOfDisplayColumn - this.numberOfDisplayLine + 1);

					UpdateScrollBar();
					UpdateDisplay();
					break;
				case Keys.Home:
					if (e.Modifiers == Keys.Control)
					{
						this.firstCharacterAbsoluteOffset = 0;

						UpdateScrollBar();
						UpdateDisplay();
					}
					break;
				case Keys.End:
					if (e.Modifiers == Keys.Control)
					{
						this.firstCharacterAbsoluteOffset = this.numberOfDisplayColumn * (this.fileStream.Length / this.numberOfDisplayColumn - this.numberOfDisplayLine + 1);

						UpdateScrollBar();
						UpdateDisplay();
					}
					break;
			}
			UpdateCursorInfo();
		}

		private void textBoxHexa_KeyUp(object sender, KeyEventArgs e)
		{
			UpdateCursorInfo();
		}

		private void textBoxHexa_Click(object sender, EventArgs e)
		{
			UpdateCursorInfo();
		}

		void textBoxHexa_MouseWheel(object sender, MouseEventArgs e)
		{
			this.firstCharacterAbsoluteOffset -= this.numberOfDisplayColumn * Math.Sign(e.Delta) * ((ModifierKeys & Keys.Control) == Keys.Control ? this.numberOfDisplayLine : 1);
			if (this.firstCharacterAbsoluteOffset < 0)
				this.firstCharacterAbsoluteOffset = 0;
			else if (this.firstCharacterAbsoluteOffset > this.lastCharacterAbsoluteOffset)
				this.firstCharacterAbsoluteOffset = this.numberOfDisplayColumn * (this.lastCharacterAbsoluteOffset / this.numberOfDisplayColumn - this.numberOfDisplayLine + 1);

			UpdateScrollBar();
			UpdateDisplay();
		}

		private void registerToolStripMenuItem_Click(object sender, EventArgs e)
		{
			FileShellExtension.Register("*", "FastHexaViewer", "Fast Hexa Viewer", string.Format("\"{0}\" \"%L\"", Application.ExecutablePath));
		}

		private void unregisterToolStripMenuItem_Click(object sender, EventArgs e)
		{
			FileShellExtension.Unregister("*", "FastHexaViewer");
		}

		private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
		{
			//MessageBox.Show("Fast Hexa Viewer\r\n    Version 0.1\r\nCopyright  2006 Rgis Cosnier\r\nhttp://regis.cosnier.free.fr");
			AboutForm aboutForm = new AboutForm();
			aboutForm.ShowDialog(this);
		}

		private long GetAbsoluteOffsetFromCursor()
		{
			int cursorPosition = this.textBoxHexa.SelectionStart;
			int lineIndex = this.textBoxHexa.GetLineFromCharIndex(cursorPosition);
			int fistCharIndex = this.textBoxHexa.GetFirstCharIndexFromLine(lineIndex);
			int columnIndex = cursorPosition - fistCharIndex;

			int relativeOffset = (columnIndex - this.numberOfAddressDigit - 3) / 3;
			if (relativeOffset >= this.numberOfDisplayColumn)
			{
				relativeOffset = (columnIndex - this.numberOfAddressDigit - 3 - this.numberOfDisplayColumn * 3 - 2);
				if (relativeOffset < 0 || relativeOffset >= this.numberOfDisplayColumn)
					relativeOffset = this.numberOfDisplayColumn - 1;
			}
			else if (relativeOffset < 0)
				relativeOffset = 0;
			return this.firstCharacterAbsoluteOffset + lineIndex * this.numberOfDisplayColumn + relativeOffset;
		}

		private void SetBinaryCursorFromAbsoluteOffset(long absoluteOffset)
		{
			long modulo = 0;
			long absoluteLineIndex0 = Math.DivRem(absoluteOffset, this.numberOfDisplayColumn, out modulo);

			long absoluteLineIndex = absoluteLineIndex0 - this.numberOfDisplayLine / 2;
			if (absoluteLineIndex < 0)
				absoluteLineIndex = 0;

			this.firstCharacterAbsoluteOffset = absoluteLineIndex * this.numberOfDisplayColumn;

			UpdateScrollBar();
			UpdateDisplay();

			int lineIndex = (int)(absoluteLineIndex0 - absoluteLineIndex);
			int columnIndex = this.numberOfAddressDigit + 3 + (int)modulo * 3;

			this.textBoxHexa.SelectionStart = lineIndex * (this.numberOfAddressDigit + 3 + this.numberOfDisplayColumn * 4 + 2 + 2) + columnIndex;
			this.textBoxHexa.SelectionLength = 2;
		}

		private void SetAsciiCursorFromAbsoluteOffset(long absoluteOffset)
		{
			long modulo = 0;
			long absoluteLineIndex0 = Math.DivRem(absoluteOffset, this.numberOfDisplayColumn, out modulo);

			long absoluteLineIndex = absoluteLineIndex0 - this.numberOfDisplayLine / 2;
			if (absoluteLineIndex < 0)
				absoluteLineIndex = 0;

			this.firstCharacterAbsoluteOffset = absoluteLineIndex * this.numberOfDisplayColumn;

			UpdateScrollBar();
			UpdateDisplay();

			int lineIndex = (int)(absoluteLineIndex0 - absoluteLineIndex);
			int columnIndex = this.numberOfAddressDigit + 3 + this.numberOfDisplayColumn * 3 + 2 + (int)modulo;

			this.textBoxHexa.SelectionStart = lineIndex * (this.numberOfAddressDigit + 3 + this.numberOfDisplayColumn * 4 + 2 + 2) + columnIndex;
			this.textBoxHexa.SelectionLength = 1;
		}

		private void goToAddressToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (this.gotoForm.ShowDialog(this) == DialogResult.OK)
			{
				if (this.gotoForm.GotoAsciiSection)
					SetAsciiCursorFromAbsoluteOffset(this.gotoForm.Address);
				else
					SetBinaryCursorFromAbsoluteOffset(this.gotoForm.Address);
			}
		}

		// Bitap algorithm (http://en.wikipedia.org/wiki/Bitap_algorithm)
		private int BitapSearch(byte[] text, byte[] pattern, int textLengh)
		{
			int result = -1;
			int m = pattern.Length;
			int i, k;

			if (pattern.Length == 0) return 0;

			// Initialize the bit array R
			bool[] R = new bool[m + 1];
			for (k = 1; k <= m; ++k)
				R[k] = false;
			R[0] = true;

			for (i = 0; i != textLengh; ++i)
			{
				// Update the bit array
				for (k = m; k >= 1; --k)
					R[k] = R[k - 1] && (text[i] == pattern[k - 1]);

				if (R[m])
				{
					result = i - m + 1;
					break;
				}
			}

			return result;
		}

		private int BitapSearch(byte[] text, char[] pattern, int textLengh, bool matchCase)
		{
			int result = -1;
			int m = pattern.Length;
			int i, k;

			if (pattern.Length == 0) return 0;

			// Initialize the bit array R
			bool[] R = new bool[m + 1];
			for (k = 1; k <= m; ++k)
				R[k] = false;
			R[0] = true;

			for (i = 0; i != textLengh; ++i)
			{
				// Update the bit array
				for (k = m; k >= 1; --k)
					R[k] = R[k - 1] &&
						(matchCase ?
							(char)text[i] == pattern[k - 1] :
							char.ToLower((char)text[i]) == char.ToLower(pattern[k - 1])
						);

				if (R[m])
				{
					result = i - m + 1;
					break;
				}
			}

			return result;
		}

		private void searchToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (this.searchForm.ShowDialog(this) == DialogResult.OK)
			{
				this.currentSearchAbsoluteOffset = this.firstCharacterAbsoluteOffset;
				Search();
			}
		}

		private void continueSearchingToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Search();
		}

		private long currentSearchAbsoluteOffset = 0;

		private void Search() //     4D 69 63 72 6F 73 6F 66 74
		{
			this.fileStream.Position = this.currentSearchAbsoluteOffset;
			this.searchResult = -1;
			ContinueSearch = true;

			this.searchTread = new Thread(new ThreadStart(delegate()
			{
				byte[] searchBuffer = new byte[4096];
				int patternLength = this.searchForm.FindAscii ? this.searchForm.SearchString.Length : this.searchForm.SearchPattern.Length;

				while (ContinueSearch)
				{
					int byteRead = 0;
					long offset = this.currentSearchAbsoluteOffset;
					lock (this)
					{
						if (this.currentSearchAbsoluteOffset < 0) this.currentSearchAbsoluteOffset = 0;
						else if (this.currentSearchAbsoluteOffset > this.fileStream.Length) break;

						this.fileStream.Position = this.currentSearchAbsoluteOffset;
						byteRead = this.fileStream.Read(searchBuffer, 0, searchBuffer.Length);
						this.currentSearchAbsoluteOffset += searchBuffer.Length - patternLength;
					}
					if (byteRead > 0)
					{
						int pos = -1;
						if (this.searchForm.FindAscii)
							pos = BitapSearch(searchBuffer, this.searchForm.SearchString.ToCharArray(), byteRead, this.searchForm.MatchCase);
						else
							pos = BitapSearch(searchBuffer, this.searchForm.SearchPattern, byteRead);
						if (pos != -1)
						{
							this.searchResult = pos + offset;
							this.BeginInvoke(new ThreadStart(delegate()
							{
								this.waitForm.Visible = false;
								ContinueSearch = false;

								SetBinaryCursorFromAbsoluteOffset(this.searchResult);
								UpdateCursorInfo();

								this.currentSearchAbsoluteOffset = this.searchResult + this.searchForm.SearchPattern.Length;
							}));
							return;
						}
					}
					else
						break;
				}

				this.BeginInvoke(new ThreadStart(delegate()
				{
					this.waitForm.Visible = false;
					ContinueSearch = false;

					lock (this)
					{
						if (this.currentSearchAbsoluteOffset > this.fileStream.Length)
							this.currentSearchAbsoluteOffset = this.fileStream.Length;

						this.firstCharacterAbsoluteOffset = (this.currentSearchAbsoluteOffset / this.numberOfDisplayColumn) * this.numberOfDisplayColumn;
					}
					UpdateScrollBar();
					UpdateDisplay();
					UpdateCursorInfo();

					MessageBox.Show(this, "Pattern not found!");
				}));

			}));
			this.searchTread.Name = "Search";
			this.searchTread.IsBackground = true;
			this.searchTread.Start();
			this.waitForm.Show(this);
		}

		void waitForm_Stopped(object sender, EventArgs e)
		{
			this.waitForm.Visible = false;
			ContinueSearch = false;

			lock (this) { this.firstCharacterAbsoluteOffset = (this.currentSearchAbsoluteOffset / this.numberOfDisplayColumn) * this.numberOfDisplayColumn; }
			UpdateScrollBar();
			UpdateDisplay();
			UpdateCursorInfo();
		}

		private void fileToolStripMenuItem_DropDownOpening(object sender, EventArgs e)
		{
			this.openToolStripMenuItem.Enabled = !ContinueSearch;
		}

		private void toolsToolStripMenuItem_DropDownOpening(object sender, EventArgs e)
		{
			this.continueSearchingToolStripMenuItem.Enabled = this.searchToolStripMenuItem.Enabled = this.goToAddressToolStripMenuItem.Enabled = !ContinueSearch;
		}

		protected override void OnPaintBackground(PaintEventArgs e)
		{
			//base.OnPaintBackground(e);
		}
	}
}